type 'a t = {elements: 'a Dynarray.t; ordre: 'a -> 'a -> bool}
(* Il vaudrait mieux faire un foncteur
   pour ne pas stocker la fonction de comparaison dans chaque tas,
   mais on a choisi la simplicit pour ce TP. *)

let length t = Dynarray.length t.elements

(* Arbres binaires *)
let pere i =
  assert (i>0);
  pred i  /2

(* let fils i = 2*i +1, 2*i +2 *)
(* Autre version pour les fous d'optimisation ou ceux qui aiment le code cryptique: *)
let fils i = succ (i lsl 1), succ i lsl 1

(* Tas *)
let rec percole t noeud =
  let arbre = t.elements in
  let valeur_noeud = Dynarray.get arbre noeud in
  let fils_g,fils_d = fils noeud in
  if fils_g < Dynarray.length arbre then (* si [noeud] a un fils gauche *)
    let valeur_fils_g = Dynarray.get arbre fils_g in
    if fils_d < Dynarray.length arbre then (* si [noeud] a un fils droit *)
      let valeur_fils_d = Dynarray.get arbre fils_d in
      (* Le plus petit des deux fils de [noeud], et sa valeur: *)
      let min_fils, valeur_min_fils =
        if t.ordre valeur_fils_g valeur_fils_d then fils_g,valeur_fils_g else fils_d,valeur_fils_d in
      (if t.ordre valeur_min_fils valeur_noeud then
         (* sinon il n'y a rien  faire : [arbre] est dj un tas *)
         (Dynarray.set arbre noeud valeur_min_fils; Dynarray.set arbre min_fils valeur_noeud;
          percole t min_fils))
    else (* si [noeud] a seulement un fils gauche *)
      (if t.ordre valeur_fils_g valeur_noeud then
         (Dynarray.set arbre noeud valeur_fils_g; Dynarray.set arbre fils_g valeur_noeud;
          percole t fils_g))
(* Si [noeud] n'a pas de fils gauche, il n'a pas non plus de fils droit. *)

let take_min t =
  let arbre = t.elements in
  let result = Dynarray.get arbre 0 in
  let new_racine = Dynarray.take arbre in
  if Dynarray.length arbre > 0 (* Pour n'appeler [Dynarray.set] que si l'arbre n'est pas vide. *)
  then (Dynarray.set arbre 0 new_racine; percole t 0);
  result

let of_array ordre array =
  let t = {elements = Dynarray.of_array array; ordre=ordre} in
  let n = Array.length array in
  if n > 1 then
    for i = pere (pred n) downto 0 do (* On commence au dernier noeud ayant un fils*)
      percole t i;
    done;
  t

let sort ordre array =
  let tas = of_array ordre array in
  Array.init (Array.length array) (fun _ -> take_min tas)
(* Remarque: on pourrait viter les redimenseionnements du tableau. *)

let echange dynarray a b =
  let temp = Dynarray.get dynarray a in
  Dynarray.set dynarray a (Dynarray.get dynarray b);
  Dynarray.set dynarray b temp

let rec percole_bottom_up t noeud =
  let arbre = t.elements in
  if noeud > 0 then
    let pere = pere noeud in
    if t.ordre (Dynarray.get arbre noeud) (Dynarray.get arbre pere) then
      (echange arbre noeud pere; percole_bottom_up t pere)

(* On cre une feuille d'tiquette [elt],
et tant que elt est plus grand que son pre on l'change avec lui,
le faisant ainsi remonter de bas en haut (bottom-up) *)
let insere elt t =
  let arbre = t.elements in
  Dynarray.add elt arbre;
  percole_bottom_up t (pred (Dynarray.length arbre))
